/////////////////////////////////////////////////////////////////////////////
//Doppelpendel ohne Reibung mit Runge-Kutta 4 - Poincaré-Schnitt zu zeichnen
/////////////////////////////////////////////////////////////////////////////

// Attribute:

float w1,w2; //Winkel von der Senkrechten weg
float W1,W2; //Für Runge-Kutta verwendet: Winkel in Ruhelage: PI
float g1,g2; //Winkelgeschwindigkeiten    
float dt = 0.01;      // Zeitschritt
float masse;
float E; //Gesamtenergie mit Nullniveau in Ruhelage
float L = 180;
float G = 0;//9.81;         // Ortfaktor für Schwerkraft
float FensterDim = 1200;
float ruhelage = FensterDim/2;         // Ruhelag in x-Koordinate
float EKontrolleA;//Gesamtenergie 
float wK,gK; 
color farbe; float f =0;
float D1,D2; //Diskriminaten für die Berechnung der fehlenden Winkelgeschwindigkeit bei gleicher Energie
int M, S,K, s; //M: Anzahl Punkte pro Startpunkt, S: Anzahl Startpunkte / Anzahl Threads, Zähler für S
int P; //nach P Folgepunkten neuen Startpunkt wählen
ArrayList<PVector> punkteW = new ArrayList<PVector>(); //Zum speichern der Kurvenpunkte

///////////////////////////////////////////////////////////////////////////

void setup() {
  size(1200,1200);
  background(0);  
  colorMode(HSB,256);
  ArrayList<MyThread> workers = new ArrayList<MyThread>();
  
  boolean finished;
  E = 1000;
  M = 5000;//Anzahl Punkte für einen Thread
  int processors = Runtime.getRuntime().availableProcessors();
  S = processors; 
  s = 0;
  if(S==1) K =1; //Nur ein Core vorhanden
  else K = S-1; //Anzahl der gestarteten Threads
  P = 5000; //nach P Folgepunkten neuer Startpunkt und neue Farbe
  println("Der Prozessor hat " + S + " Kerne"); 
  println(K + " Threads werden gestartet"); 
  
    while(s<K){ //K mal starten ...
      println(s);
      for (int i=0; i<K; i++) { //K Threads
        MyThread t = new MyThread(E,M,P);
        // nun starte den Thread:
        t.start(); 
        workers.add(t); //Es werden 16 Threads gestartet   
       }
             
       do  {
        // prüfe, ob alle Prozesse schon fertig sind
        finished = true;
        for (MyThread worker : workers) 
            if (worker.isAlive()) finished = false;  
              
         // warte ein wenig, damit nicht zu viele Nachfragen
         // den Rechner blockieren
        delay(10);
      } while (!finished);
      
      
       Thread.currentThread().interrupt();
       for (MyThread worker : workers){
          punkteW = worker.punkte;
          punkteZeichnen();         
       }
       s++;  
    }//end of while  
   
}


/////////////////////////////////////////////////////////
void draw() { 
 
  
}

void punkteZeichnen(){  
    f = (f+1) % 256;
    farbe = color(f,256,256);
    stroke(farbe);;
    strokeWeight(1);
    for(int i =0; i< punkteW.size()-1 ;i++){
        if(i % P ==0)farbe = color(random(256),random(256),random(256));
        point(punkteW.get(i).x,punkteW.get(i).y);       
     }  
    punkteW.clear();
}


////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
public class MyThread extends Thread {
  int n = 4;            // Anzahl der Variablen
  float E;
  int M;
  int P,p;
  int m; 
  float w1,w2,W1,W2;
  float g1,g2;
  float[] ZVaktuell ; // Initialisierung w1,w2,g1,g2
  ArrayList<PVector> punkte ;
  boolean fertig;
  float wVor,wNach;
  
  // Thread anlegen und Parameter übergeben
  public MyThread(float E, int M, int P){
     this.E = E;
     this.M = M;    
     this.P = P;
     fertig = false;
     masse =1;
     m = 0;
     p = 0;
     ZVaktuell = new float[4];
     punkte = new ArrayList<PVector>();
  }
/////////////////////////////////////////////////////////////
  public void run(){
    neuStarten(); 
    while(!fertig){
    folgePunkteBerechnen();      
    } 
   
  
 }
 
//////////////////////////////////////////////////////////
float[] ZVaktualisieren(float[] Y) {
    float[] ZustandsVektor = new float[n]; 
    ZustandsVektor[0]=Y[2];
    ZustandsVektor[1]=Y[3];
    // benötigte Werte :
    float sinDiff = sin(Y[0]-Y[1]);
    float cosDiff = cos(Y[0]-Y[1]);
    float cosProdukt = cosDiff * cosDiff;
    float s0 = sin(Y[0]);
    float s1 = sin(Y[1]);
    float alpha = - Y[3]*Y[3]*sinDiff + 2*G/L*s0 ;
    float beta = Y[2]*Y[2]*sinDiff + G/L*s1 ;
    ZustandsVektor[2]= 1 / (2-cosProdukt) * (alpha - cosDiff * beta); //Geschwindigkeit 1
    ZustandsVektor[3]= 1 / (2-cosProdukt) * (2*beta - cosDiff * alpha);//Geschwindigkeit 2 
    return ZustandsVektor;
}
////////////////////////////////////////////////////////////////////////////
void pendelUpdate(){
    // Zustand nach nächstem Zeitschritt mit Runge-Kutta 4 berechnen
    // "float[] ZVaktualisieren" Methode der zeitlichen Ableitung dy/dt (4-dimensional). 
    // Die DGL hängt hier nicht explizit von t ab
 
    //Die folgende Zeile ist Bestandteil für das Zeichnen eines Poincaré-Schnitts   
     wVor = ZVaktuell[1];
    float[] K1 = ZVaktualisieren(ZVaktuell); 
    // K2 =ZustandsVektor(t0 + dt/2, ZVaktuell + dt/2 * K1)
    float[] ZVtemp = new float[n];
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K1[i]*dt*0.5;    
    }
    float[] K2 = ZVaktualisieren(ZVtemp);
    // K3 =ZustandsVektor(t0 + dt/2, ZVaktuell + dt/2 * K2)
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K2[i]*dt*0.5;
    }
    float[] K3 = ZVaktualisieren(ZVtemp);
    // K4 =ZustandsVektor(t0 + dt, ZVaktuell + dt * K3)
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K3[i]*dt;
    }
    float[] K4 = ZVaktualisieren(ZVtemp);
    // Zusammenfassung
    //  ZVaktuell = ZVaktuell + dt/6 * (K1 + 2*(K2+K3) + K4)
    for (int i=0; i<n; i++) {
      ZVaktuell[i] = ZVaktuell[i] + dt / 6.0 * (K1[i] + 2.0*(K2[i] + K3[i]) + K4[i]);
    }  
    //Die folgenden Zeilen sind Bestandteil für das Zeichnen eines Poincaré-Schnitts
     wNach = ZVaktuell[1];
     if(wVor < PI && wNach > PI ){
         if(p % P ==0) neuStarten();
         wK = map(ZVaktuell[0],0,2*PI,0,FensterDim);  //Koordinaten der Poincaré-Schnitt Punkte 
         gK = map(ZVaktuell[2],-0.35, 0.35,FensterDim,0);   
         punkte.add(new PVector(wK,gK));
         m++;  
         p++;
      }    
    //Winkel immer zwischen 0 und 2*PI 
    if(ZVaktuell[0] > 2*PI) ZVaktuell[0] = ZVaktuell[0] - 2*PI;
    if(ZVaktuell[0] < 0)    ZVaktuell[0] = ZVaktuell[0] + 2*PI;
    if(ZVaktuell[1] > 2*PI) ZVaktuell[1] = ZVaktuell[1] - 2*PI;
    if(ZVaktuell[1] < 0)    ZVaktuell[1] = ZVaktuell[1] + 2*PI;
    W1 = ZVaktuell[0]; w1 = PI-W1;
    W2 = ZVaktuell[1]; w2 = PI-W2;
    g1= ZVaktuell[2];
    g2 = ZVaktuell[3]; 
 }
 
  void folgePunkteBerechnen(){ 
    pendelUpdate();
    if(m >= M){
      fertig = true; 
    }   
   }

 ///////////////////////////////////////////////////////////////////////////////
  void neuStarten(){
    
    w1=random(-0.1,0.1);   
    W1 = PI-w1;
    w2= random(-0.1,0.1);
    W2 = PI-w2;
    //Diese Konstruktion, einmal die Winkelgeschwindigkeit g1 und ein andermal g2 zufällig zu bestimmen
    //dient dazu, dass mehrere verschieden Startpunkte im Phasendiagramm verwendet werden.
    if(random(0,1)<0.5){
    g2=random(-0.2,0.2); 
    g1 = berechneG1();
    }else{  
    g1=random(-0.2,0.2); 
    g2 = berechneG2();
    }
    ZVaktuell[0] = W1;
    ZVaktuell[1] = W2;         
    ZVaktuell[2] = g1;
    ZVaktuell[3] = g2; 
    //m = 0;
  }  
//////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
  float berechneG1(){//Damit Energie konstant ist  
    D1 = g2*g2*cos(w2-w1)*cos(w2-w1)/4 -g2*g2/2+ G*(2*cos(w1)+cos(w2)-3)/L + E/(L*L*masse);
    if(D1>=0){
      g1 = -g2*cos(w1-w2)/2 + sqrt(D1);  
    }else {
      println("Größe der Energie oder Größe der zufällig bestimmten Werte ändern!");
    }
    return g1; 
  }
////////////////////////////////////////////////////////////////////
  float berechneG2(){//Damit Energie konstant ist 
    D2 = g1*g1*cos(w2-w1)*cos(w2-w1) -2*g1*g1+ 2*G*(2*cos(w1)+cos(w2)-3)/L + 2*E/(L*L*masse);
    if(D2>=0){
       g2 = -g1*cos(w1-w2) + sqrt(D2);
    }else{
       println("Größe der Energie oder Größe der zufällig bestimmten Werte ändern!");
    }
    return g2; 
  }
//////////////////////////////////////////////////////////////////////////
  void mousePressed() { 
    
  }
///////////////////////////////////////////////////////////////////////////////
}
